# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import multiprocessing
from multiprocessing.pool import Pool
from random import uniform
import logging
# logging.basicConfig(filename='suchpy_log.txt', level=logging.DEBUG,
#  format='%(asctime)s - %(levelname)s - %(message)s')


from supplychainpy.demand._forecast_demand import Forecast
from supplychainpy.demand.regression import LinearRegression

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class Verbose:
    def __init__(self, parents: list=None, standard_error: float=None, smoothing_level: float=None):
        print('inside __init__')
        self.arg1 = parents
        self.arg2 = standard_error
        self.arg3 = smoothing_level

    def __call__(self, f):
        print('Inside __call__')

        def annotate(*args):
            for i in args:
                print('inside methods {}'.format(i))
            f(*args)

        return annotate()


class Individual:
    _genome = 0

    def __init__(
            self, name: str='offspring', overide: bool=False, gene_count: int=12,
            forecast_type: str='ses'
        ):

        self._name = name
        self._gene_count = gene_count
        self._forecast_type = forecast_type
        if not overide:
            self._genome = self.__genome_generator()

    def __repr__(self):
        return '{}: {}'.format(self._name, self._genome)

    @property
    def gene_count(self) -> int:
        return self.gene_count

    @gene_count.setter
    def gene_count(self, value: int):
        self._gene_count = value

    @property
    def genome(self):
        return self._genome

    @genome.setter
    def genome(self, val: tuple):
        self._genome = val

    @property
    def name(self):
        return self._name

    def __genome_generator(self):
        genome = ()
        if self._forecast_type == 'ses':
            genome = tuple([uniform(0, 1) for i in range(0, self._gene_count)])
        else:
            genome = tuple([(uniform(0, 1), uniform(0, 1)) for i in range(0, self._gene_count)])

        return genome


class Population:
    """ Create a population of individuals to reproduce. """

    __slots__ = ['individuals', 'mutation_probability']
    _recombination_type = ('single_point', 'two_point', 'uniform')

    def __init__(self, individuals: list, mutation_probability: float = 0.2):
        self.individuals = individuals
        self.mutation_probability = mutation_probability

    def reproduce(self, recombination_type: str = 'single_point'):
        """ Coordinates the reproduction of two individuals, using one of three recombination methods 'single_point,
        two_point or uniform'.

        Args:
            recombination_type: The method of recombination used to

        Yields:
            (dict):     The result of recombination and mutation.

        """
        log.log(logging.INFO, '{} reproduction started'.format(recombination_type))

        if recombination_type == self._recombination_type[0]:
            log.log(logging.INFO, 'started single point recombination')
            yield [i for i in self._single_point_crossover_recombination()][0]
        if recombination_type == self._recombination_type[1]:
            log.log(logging.INFO, 'started two point recombination')
            yield [i for i in self._two_point_crossover_recombination()][0]
        if recombination_type == self._recombination_type[2]:
            log.log(logging.INFO, 'logging.INFO,started uniform recombination')
            yield [i for i in self._uniform_crossover_recombination()][0]

    def _single_point_crossover_recombination(self) -> object:
        """ Selects genes from the parent to crossover based on one point along the chromosome.

        Yields:
            individual (dict):          Parents p1-xxxxxxxxxxxx and p2-oooooooooooo will produce offspring c1-xxxxxxooooo
                                        and c2-ooooooxxxxxx. Individual yields the first offspring (c1).
            individual_two (dict):      Parents p1-xxxxxxxxxxxxx and p2-oooooooooooo will produce offspring c1-xxxxxxooooo
                                        and c2-ooooooxxxxxx. Individual_two yields the second offspring (c2).
        """

        genome_count = 0
        new_individual = {}
        new_individual_two = {}

        mutation_count = 0
        mutation_index = 0

        try:

            if len(self.individuals) > 1:
                for index, individual in enumerate(self.individuals):
                    for key, value in individual.items():
                        population_allele_count = len(self.individuals) * len(individual.items())
                        number_of_mutations_allowed = round(population_allele_count * self.mutation_probability)
                        mutation_index += 1
                        if (genome_count <= 5 and len(new_individual) < 12) or (
                                            18 >= genome_count > 12 > len(new_individual)):

                            if mutation_index == number_of_mutations_allowed:
                                new_individual.update(self._mutation({key: value}))
                                genome_count += 1
                                mutation_index = 0
                                mutation_count += 1
                            else:
                                new_individual.update({key: value})
                                genome_count += 1

                        elif (genome_count >= 6 and len(new_individual_two) < 12) or (18 < genome_count <= 24 and len(
                                new_individual) < 12):

                            if mutation_index == number_of_mutations_allowed:
                                new_individual_two.update(self._mutation({key: value}))
                                genome_count += 1
                                mutation_index = 0
                                mutation_count += 1
                            else:
                                new_individual.update({key: value})
                                genome_count += 1

                        else:
                            genome_count += 1

                        if genome_count == 24:
                            genome_count = 0
                            # if new_individual != new_individual_two:
                            # log.debug('single point crossover was successful')
                            # else:
                            # log.debug('single point crossover resulted in clone of parent')
                            yield (new_individual)
                            yield (new_individual_two)

            else:
                raise ValueError()

        except ValueError:
            print('Population Size is too small for effective genetic recombination and reproduction of offspring')

    def _two_point_crossover_recombination(self):
        """ Selects genes from the parent to crossover based on two points along the chromosome.

        Yields:
            individual (dict):          Parents p1-xxxxxxxxxxxxx and p2-oooooooooooo will produce offspring c1-xxxooooooxxx
                                        and c2-oooxxxxxxooo. Individual yields the first offspring (c1).
            individual_two (dict):      Parents p1-xxxxxxxxxxxxx and p2-oooooooooooo will produce offspring c1-xxxooooooxxx
                                        and c2-oooxxxxxxooo. Individual_two yields the second offspring (c2).
        """

        genome_count = 0
        new_individual = {}

        mutation_count = 0
        mutation_index = 0
        new_individual_two = {}

        try:
            if len(self.individuals) > 1:
                for index, individual in enumerate(self.individuals):
                    for key, value in individual.items():
                        population_allele_count = len(self.individuals) * len(individual.items())
                        number_of_mutations_allowed = round(population_allele_count * self.mutation_probability)
                        mutation_index += 1
                        if (2 >= genome_count and len(new_individual) < 12) or (15 <= genome_count <= 20) or (
                                        9 <= genome_count <= 11):
                            if mutation_index == number_of_mutations_allowed:
                                new_individual.update(self._mutation({key: value}))
                                genome_count += 1
                                mutation_index = 0
                                mutation_count += 1
                            else:
                                new_individual.update({key: value})
                                genome_count += 1

                        elif (12 <= genome_count <= 14 and len(new_individual_two) < 12) or (
                                        3 <= genome_count <= 8) and (len(
                            new_individual) < 12 or 21 <= genome_count <= 23):
                            if mutation_index == number_of_mutations_allowed:
                                new_individual.update(self._mutation({key: value}))
                                genome_count += 1
                                mutation_index = 0
                                mutation_count += 1
                            else:
                                new_individual.update({key: value})
                                genome_count += 1
                        else:
                            genome_count += 1

                        if genome_count == 24:
                            genome_count = 0
                            # if new_individual != new_individual_two:
                            # log.debug('single point crossover was successful')
                            # else:
                            # log.debug('single point crossover resulted in clone of parent')
                            yield (new_individual)
                            yield (new_individual_two)



        except OSError as e:

            print(e)

    # TODO-feature uniform crossover recombination for reproduction
    def _uniform_crossover_recombination(self):
        pass

    @staticmethod
    def _mutation(gene: dict):
        """Mutates a gene and reinserts it into the chromosome. A phenotype expressed as a standard error is altered
        by having the bit representing it flipped.

        Args:
            gene:   The alpha value and bit representing the positive or negative fit of the gene.

        Returns:
            dict:   The new gene with an altered value for the it representing the positive or negative fit of the gene.

        """

        original_gene = [[k, v] for k, v in gene.items()]

        def mutate():

            if original_gene[0][1] == 0:
                log.debug(
                    'A mutation occurred on gene {}. The result of the mutation is {}'.format({original_gene[0][0]: 0},
                                                                                              {original_gene[0][0]: 1}))
                return {original_gene[0][0]: 1}
            elif original_gene[0][1] == 1:
                log.debug(
                    'A mutation occurred on gene {}. The result of the mutation is {}'.format({original_gene[0][0]: 1},
                                                                                              {original_gene[0][0]: 0}))
                return {original_gene[0][0]: 0}
            else:
                return {original_gene[0][0]: original_gene[0][0]}

        return mutate()


class DiversifyPopulation(Population):
    def __init__(self, individuals: list):
        super().__init__(individuals)


class OptimiseSmoothingLevelGeneticAlgorithm:
    __generation = 0

    def __str__(self):
        return "{} {}".format(type(self), id(self) )

    def __init__(self, orders: list, **kwargs):
        self.__orders = orders
        if len(kwargs) != 0:
            self.__average_order = kwargs['average_order']
            self.__population_size = kwargs['population_size']
            self.__standard_error = kwargs['standard_error']
            self.__recombination_type = kwargs['recombination_type']
            log.log(logging.DEBUG, "Initialising GA solver for {}.".format(self.__str__()))

    def initial_population(self, individual_type: str = 'ses'):
        """Initialises population and initiates the optimisation of the standard error by searching for an optimum
        alpha value.
        Returns:

        """
        log.log(logging.INFO, "Initialising population.")
        return self._initialise_smoothing_level_evolutionary_algorithm_population(individual_type=individual_type)

    @property
    def population_size(self):
        return self.__population_size

    @population_size.setter
    def population_size(self, population_size):
        self.__population_size = population_size

    def _initialise_smoothing_level_evolutionary_algorithm_population(self, individual_type: str):
        """ Starts the process for creating the population. The number of parents is specified during the
         initialisation of the class. """

        parents = []
        parents_population = []

        while len(parents_population) < self.__population_size:
            for i in range(0, self.__population_size):
                parent = Individual(name='parent', forecast_type=individual_type)
                log.debug('Initial parent created {}'.format(parent))
                parents.append(parent)

            populations_genome = [i for i in self.generate_smoothing_level_genome(population=parents,
                                                                                  individual_type=individual_type)]
            log.debug('Population with genome {}'.format(populations_genome))
            populations_traits = [i for i in self.express_smoothing_level_genome(individuals_genome=populations_genome,
                                                                                 standard_error=self.__standard_error)]

            fit_population = [i for i in
                              self._population_fitness(population=populations_traits, individual_type=individual_type)]

            log.debug('Fit population {}'.format(fit_population))

            parents_population += fit_population

        create_offspring = Population(individuals=parents_population)

        # population reproduce
        new_population = [i for i in create_offspring.reproduce(recombination_type=self.__recombination_type)]

        if new_population is None:
            return 0

        parent_offspring_population = []
        new_individuals = []
        while len(new_population) < self.__population_size * 5:
            for po in new_population:
                pke = po.keys()
                parent_offspring_population.append(tuple(pke))

            for genome in parent_offspring_population:
                new_individual = Individual(overide=True)
                new_individual.genome = genome
                new_individuals.append(new_individual)

            # while population allele boundary ie 50 70 95 is less than specified number.
            new_population_genome = [i for i in self.generate_smoothing_level_genome(population=new_individuals,
                                                                                     individual_type=individual_type)]

            new_populations_traits = [i for i in
                                      self.express_smoothing_level_genome(individuals_genome=new_population_genome,
                                                                          standard_error=self.__standard_error)]

            new_fit_population = [i for i in self._population_fitness(population=new_populations_traits,
                                                                      individual_type=individual_type)]
            new_population = new_fit_population

        new_individuals.clear()

        new_individuals = [i for i in self.create_individuals(new_population)]

        final_error = [i for i in self.generate_smoothing_level_genome(population=new_individuals,
                                                                       individual_type=individual_type)]

        minimum_smoothing_level = min(zip(final_error[0].values(), final_error[0].keys()))

        return minimum_smoothing_level

    @staticmethod
    def create_individuals(new_population: list) -> list:
        """Create individuals using class from genomes striped during processing fitness.
        Args:
            new_population (list):  new population of individual genomes.
        """
        parent_offspring_population = []

        for po in new_population:
            pke = po.keys()
            parent_offspring_population.append(tuple(pke))

        for genome in parent_offspring_population:
            new_individual = Individual(overide=True)
            new_individual.genome = genome
            yield new_individual

    @staticmethod
    def _population_fitness(population: list, individual_type: str = 'ses') -> list:
        """ Assess the population for fitness before crossover and creating next generation. Positive traits
        should be reflected by more than 70% of the genes in the genome.

        Args:
            population (list): A population with expressed traits (phenotypes). The standard_errors representing
            the genome of the individual have been calculated and expressed as one or zero if below or above
            the original standard error respectively.

        Returns:
            fit_population (list):  A population of individuals with a probability of procreating above 70% for simple
                                    exponential smoothing and 30% for holts trend correcting. """

        for individual in population:
            procreation_probability = sum(individual.values()) / len(individual.values())
            if individual_type == 'ses':
                if procreation_probability >= 0.5:
                    yield individual
            else:
                if procreation_probability >= 0.3:
                    yield individual

    def generate_smoothing_level_genome(self, population: list, individual_type: str = 'ses'):

        # stack parents and offspring into a list for next steps
        if individual_type == 'ses':
            for parent in population:
                individuals_genome = self._run_exponential_smoothing_forecast(parent.genome)
                # print(individuals_genome)
                # individuals_traits = self._express_trait(standard_error, smoothing_level, individuals_genome)

                yield individuals_genome
        else:
            for parent in population:
                individuals_genome = self._run_holts_trend_corrected_exponential_smoothing(parent.genome)
                # print(individuals_genome)
                # individuals_traits = self._express_trait(standard_error, smoothing_level, individuals_genome)

                yield individuals_genome

    @staticmethod
    def htce_para(forecast, alpha, gamma, intercept, slope):
        return [i for i in forecast.holts_trend_corrected_exponential_smoothing(alpha=alpha, gamma=gamma,
                                                                                intercept=intercept,
                                                                                slope=slope)]


    def sum_squared_error_para(self, forecast, holts_trend_corrected_smoothing, alpha, gamma, smoothing_level):
        appraised_individual = {}
        sum_squared_error = forecast.sum_squared_errors_indi_htces(squared_error=holts_trend_corrected_smoothing,
                                                            alpha=alpha, gamma=gamma)
        standard_error = forecast.standard_error(sum_squared_error, len(self.__orders), smoothing_level)
        appraised_individual.update({smoothing_level: standard_error})
        return appraised_individual

    def _run_holts_trend_corrected_exponential_smoothing(self, individual: tuple):
        f = Forecast(self.__orders, self.__average_order)
        holts_trend_corrected_smoothing = []
        demand = [{'t': index, 'demand': order} for index, order in enumerate(self.__orders, 1)]
        stats = LinearRegression(demand)
        log_stats = stats.least_squared_error(slice_end=6)

        for sm_lvl in individual:
            p = [i for i in f.holts_trend_corrected_exponential_smoothing(alpha=sm_lvl[0], gamma=sm_lvl[1],
                                                                          intercept=log_stats.get('intercept'),
                                                                          slope=log_stats.get('slope'))]
            # print(p)
            holts_trend_corrected_smoothing.append(p)

        appraised_individual = {}
        for smoothing_level in individual:
            sum_squared_error = f.sum_squared_errors_indi_htces(squared_error=holts_trend_corrected_smoothing,
                                                                alpha=smoothing_level[0], gamma=smoothing_level[1])
            standard_error = f.standard_error(sum_squared_error, len(self.__orders), smoothing_level)
            appraised_individual.update({smoothing_level: standard_error})
        # print('The standard error as a trait has been calculated {}'.format(appraised_individual))
        return appraised_individual

    def express_smoothing_level_genome(self, individuals_genome: list, standard_error):

        # stack parents and offspring into a list for next steps
        for genome in individuals_genome:
            individuals_traits = self._express_trait(standard_error, genome)
            yield individuals_traits

    def _run_exponential_smoothing_forecast(self, individual: tuple) -> dict:

        f = Forecast(self.__orders, self.__average_order)
        simple_expo_smoothing = []
        for sm_lvl in individual:
            p = [i for i in f.simple_exponential_smoothing(sm_lvl)]
            # print(p)
            simple_expo_smoothing.append(p)

        appraised_individual = {}
        for smoothing_level in individual:
            sum_squared_error = f.sum_squared_errors_indi(simple_expo_smoothing, smoothing_level)
            standard_error = f.standard_error(sum_squared_error, len(self.__orders), smoothing_level)
            appraised_individual.update({smoothing_level: standard_error})
        # print('The standard error as a trait has been calculated {}'.format(appraised_individual))
        return appraised_individual

    @staticmethod
    def _express_trait(original_standard_error: float, appraised_individual: dict):
        # fitness test? over 50% of the alleles must be positive traits. give percentage score
        # print(original_standard_error)
        for key in appraised_individual:
            if appraised_individual[key] < original_standard_error:
                appraised_individual[key] = 1
            else:
                appraised_individual[key] = 0
        # print('The traits have been verified {}'.format(appraised_individual))
        # check fitness of individual to procreate. 50% of traits need to be better smoothing_levels than the original
        return appraised_individual

    def _selection_population(self, individuals_fitness: dict, appraised_individual: list):
        pass

    def simple_exponential_smoothing_evo(self, smoothing_level_constant: float, initial_estimate_period: int,
                                         recombination_type: str = 'single_point', population_size: int = 10,
                                         forecast_length: int = 5) -> dict:
        """ Simple exponential smoothing using evolutionary algorithm for optimising smoothing level constant (alpha value)

            Args:
                initial_estimate_period (int):      The number of previous data points required for initial level estimate.
                smoothing_level_constant (float):   Best guess at smoothing level constant appropriate for forecast.

           Returns:
               dict:

            Example:

        """
        log.log(logging.INFO, "Executing simple exponential smoothing. "
                              "SMOOTHING_LEVEL: {} "
                              "INITIAL_ESTIMATE_PERIOD: {} "
                              "RECOMBINATION_TYPE: {} "
                              "POPULATION_SIZE: {} "
                              "FORECAST_LENGTH: {}".format(smoothing_level_constant,
                                                             initial_estimate_period,
                                                             recombination_type,
                                                             population_size,
                                                             forecast_length
                                                             ))
        if None != self.__recombination_type:
            recombination_type = self.__recombination_type

        sum_orders = 0

        for demand in self.__orders[:initial_estimate_period]:
            sum_orders += demand

        avg_orders = sum_orders / initial_estimate_period

        forecast_demand = Forecast(self.__orders, avg_orders)

        #calls simple_exponential_smoothing method from Forecast object
        ses_forecast = [i for i in forecast_demand.simple_exponential_smoothing(*(smoothing_level_constant,))]

        sum_squared_error = forecast_demand.sum_squared_errors(ses_forecast, smoothing_level_constant)

        standard_error = forecast_demand.standard_error(sum_squared_error, len(self.__orders),
                                                        smoothing_level_constant)

        evo_mod = OptimiseSmoothingLevelGeneticAlgorithm(orders=self.__orders,
                                                         average_order=avg_orders,
                                                         smoothing_level=smoothing_level_constant,
                                                         population_size=population_size,
                                                         standard_error=standard_error,
                                                         recombination_type=recombination_type)

        optimal_alpha = evo_mod.initial_population()

        optimal_ses_forecast = [i for i in forecast_demand.simple_exponential_smoothing(optimal_alpha[1])]

        ape = LinearRegression(optimal_ses_forecast)
        mape = forecast_demand.mean_aboslute_percentage_error_opt(optimal_ses_forecast)
        stats = ape.least_squared_error()
        simple_forecast = forecast_demand.simple_exponential_smoothing_forecast(forecast=optimal_ses_forecast,
                                                                                forecast_length=forecast_length)

        sum_squared_error = forecast_demand.sum_squared_errors(optimal_ses_forecast, optimal_alpha[1])
        standard_error = forecast_demand.standard_error(sum_squared_error, len(self.__orders),
                                                        optimal_alpha[1])
        regression = {
            'regression': [(stats.get('slope') * i) + stats.get('intercept') for i in range(0, 12)]}

        log.log(logging.INFO,
                "An OPTIMISED simple exponential smoothing forecast has been completed")
        return {'forecast_breakdown': optimal_ses_forecast, 'mape': mape, 'statistics': stats,
                'forecast': simple_forecast, 'optimal_alpha': optimal_alpha[1], 'standard_error': standard_error,
                'regression': [i for i in regression.get('regression')]}



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from decimal import Decimal


class MonteCarloWindow:
    """ Attributes for monte carlo calculations, for building the simulation window

    """

    def __init__(self):
        self._index = 0
        self._sku_id = ""
        self._lead_time = 0
        self._opening_stock = 0
        self._demand = 0
        self._closing_stock = 0
        self._backlog = 0.00
        self._holding_cost = Decimal('0.0')
        self._shortage_cost = Decimal('0.0')
        self._po_raised_flag = False
        self._purchase_order_raised_qty = 0
        self._purchase_order_receipt_qty = 0
        self._backlog_at_po_placement = 0
        self._quantity_sold = 0
        self._position = 0
        self._shortage_units = 0
        self._po_number_raised = ''
        self._po_number_received = ''
        self._revenue = 0.00
        self._sold = 0
        self._previous_backlog = 0.0

    @property
    def previous_backlog(self):
        return self._previous_backlog

    @previous_backlog.setter
    def previous_backlog(self, previous_backlog):
        self._previous_backlog = previous_backlog

    @property
    def sold(self):
        return self._sold

    @sold.setter
    def sold(self, sold):
        self._sold = sold

    @property
    def revenue(self) -> str:
        return self._revenue

    @revenue.setter
    def revenue(self, revenue):
        self._revenue = revenue

    @property
    def po_number_raised(self) -> str:
        return self._po_number_raised

    @po_number_raised.setter
    def po_number_raised(self, po_raised: str):
        self._po_number_raised = po_raised

    @property
    def po_number_received(self) -> int:
        return self._po_number_received

    @po_number_received.setter
    def po_number_received(self, po_received: int):
        self._po_number_received = po_received

    @property
    def index(self) -> int:
        return self._index

    @index.setter
    def index(self, index: int):
        self._index = index

    @property
    def shortage_units(self) -> int:
        return self._shortage_cost

    @shortage_units.setter
    def shortage_units(self, shortage_units):
        self._shortage_cost = shortage_units

    @property
    def sku_id(self) -> str:
        return self._sku_id

    @sku_id.setter
    def sku_id(self, sku_id: str):
        self._sku_id = sku_id

    @property
    def lead_time(self) -> int:
        return self._lead_time

    @lead_time.setter
    def lead_time(self, lead_time):
        self._lead_time = lead_time

    @property
    def opening_stock(self) -> int:
        return self._opening_stock

    @opening_stock.setter
    def opening_stock(self, opening_stock: int):
        self._opening_stock = opening_stock

    @property
    def demand(self) -> int:
        return self._demand

    @demand.setter
    def demand(self, demand: int):
        self._demand = demand

    @property
    def closing_stock(self) -> int:
        return self._closing_stock

    @closing_stock.setter
    def closing_stock(self, closing_stock: int):
        self._closing_stock = closing_stock

    @property
    def backlog(self) -> Decimal:
        return self._backlog

    @backlog.setter
    def backlog(self, backlog: Decimal):
        self._backlog = backlog

    @property
    def holding_cost(self) -> Decimal:
        return self._holding_cost

    @holding_cost.setter
    def holding_cost(self, holding_cost: Decimal):
        pass

    @property
    def shortage_cost(self) -> Decimal:
        return self._shortage_cost

    @shortage_cost.setter
    def shortage_cost(self, shortage_cost: Decimal):
        self._shortage_cost = shortage_cost

    @property
    def po_raised_flag(self) -> bool:
        return self._po_raised_flag

    @po_raised_flag.setter
    def po_raised_flag(self, po_raised_flag: bool):
        self._po_raised_flag = po_raised_flag

    @property
    def purchase_order_raised_qty(self) -> Decimal:
        return self._purchase_order_raised_qty

    @purchase_order_raised_qty.setter
    def purchase_order_raised_qty(self, po_raised_qty: Decimal):
        self._purchase_order_raised_qty = po_raised_qty

    @property
    def purchase_order_receipt_qty(self) -> int:
        return self._purchase_order_receipt_qty

    @purchase_order_receipt_qty.setter
    def purchase_order_receipt_qty(self, po_receipt_qty: int):
        self._purchase_order_receipt_qty = po_receipt_qty

    @property
    def backlog_at_po_placement(self) -> int:
        return self._backlog_at_po_placement

    @backlog_at_po_placement.setter
    def backlog_at_po_placement(self, backlog: int):
        self._backlog_at_po_placement = backlog

    @property
    def quantity_sold(self) -> int:
        return self._quantity_sold

    @quantity_sold.setter
    def quantity_sold(self, quantity_sold: int):
        self._quantity_sold = quantity_sold

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position: int):
        self._position = position

    def __del__(self):
        self._index = 0
        self._sku_id = ""
        self._lead_time = 0
        self._opening_stock = 0
        self._demand = 0
        self._closing_stock = 0
        self._backlog = 0.00
        self._holding_cost = Decimal('0.0')
        self._shortage_cost = Decimal('0.0')
        self._po_raised_flag = False
        self._purchase_order_raised_qty = 0
        self._purchase_order_receipt_qty = 0
        self._backlog_at_po_placement = 0
        self._quantity_sold = 0
        self._position = 0
        self._shortage_units = 0
        self._po_number_raised = ''
        self._po_number_received = ''
        self._revenue = 0.00



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

class MonteCarloFrameSummary:
    def __init__(self):
        pass

    @staticmethod
    def closing_stockout_percentage(closing_stock: list, period_length: int):
        """ Calculates the percentage of stock out that occurred during the period specified.


        Args:
            period_length (int):    length of window e.g. 12 weeks for a quarter etc.
            closing_stock (list):   list of closing stock values for a given sku over the same time frame
                                    as the period_length


        Returns:
            float:   Percentage of final closing stock values that result in stock out and backlog.

        Raises:
            ValueError: The number of stock positions and the period length must match exactly.
        """
        if len(closing_stock) != period_length:
            raise ValueError(" The number of stock positions and the period length must match exactly.\nThe"
                             " number of stock positions passed for closing stock is currently {}. "
                             "The length specified for period_length is currently {}".format(len(closing_stock),
                                                                                                 period_length))

        closing_stock_count = ([x for x in closing_stock if x <= 0])
        percentage = len(closing_stock_count) / period_length
        return percentage



import logging
from supplychainpy.simulations.simulation_window import MonteCarloWindow

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

cdef revenue(double unit_cost, double units_sold):
    return unit_cost * units_sold

cdef closing_stock(double backlog,  double opening_stock, double deliveries, double orders):
    if  (opening_stock - orders) + (deliveries) > 0:
        return ((opening_stock - orders) + deliveries) - backlog
    else:
        return 0

cdef backlog(double opening_stock, double deliveries, double demand):
    if ((opening_stock + deliveries) - demand) < 0 :
        return abs(opening_stock + deliveries) - demand
    else:
        return 0

cdef holding_cost(double closing_stock, double unit_cost, double holding_cost_percentage):
    if closing_stock > 0:
        return closing_stock * unit_cost * holding_cost_percentage
    else:
        return 0

cdef shortages(double opening_stock, double orders, double deliveries):
    if ((opening_stock - orders) + deliveries) < 0:
        return ((opening_stock - orders) + deliveries)
    else:
        return 0

cdef shortage_cost(double closing_stock, double unit_cost, double shortage_cost_percentage):
    if closing_stock > 0 :
        return closing_stock * (unit_cost * shortage_cost_percentage)
    else:
        return 0


cdef raise_po(double reorder_level, double closing_stock):
    return True if closing_stock <= reorder_level else False

cdef po_qty(double eoq, double reorder_level, double backlog, double closing_stock):
    if ((eoq + backlog + reorder_level) - closing_stock) > 0:
        return ((eoq + backlog + reorder_level) - closing_stock)
    else:
        return 0

cdef opening_stock(double reorder_level, double safety_stock, int position, double previous_closing_stock):
    if position == 1:
        stock = reorder_level #calculated ltd until put into analyse orders
    else:
        stock = previous_closing_stock
    return stock

cdef units_sold(double backlog, double opening_stock, double delivery, double demand):

    # check if opening_stock + closing_stock = 0
    if int(opening_stock) + int(delivery) == 0:
        sold = 0.00
    else:
        sold = (opening_stock + delivery) - demand - backlog

    if sold < 0:
        units_sold = demand + backlog - abs(sold)
    else:
        units_sold = sold

    return units_sold

cdef previous_backlog(double backlog, double closing_stock, double previous_backlog):
        if closing_stock == 0.00:
            previous_backlog += backlog
            return previous_backlog
        else:
            return 0

def simulation_window(list random_normal_demand, int period_length, list analysed_orders,
                     double holding_cost_percentage = 0.48,
                     double shortage_cost_percentage = 0.3):


    for sku in analysed_orders:
        if period_length != len(random_normal_demand[0][sku.sku_id]):
            raise ValueError("The period_length is currently {} and the actual length of the demand is {}. "
                             "Please make sure that the two values are equal".format(period_length,
                                                                                     len(random_normal_demand[0][
                                                                                             sku.sku_id])))

    cdef list sim_frame_collection = []
    cdef int index_item = 1
    cdef int period = 1
    cdef dict order_receipt_index
    cdef double final_stock = 0.0
    cdef dict sim_window_collection = {}
    cdef int po_qty_raised
    cdef double sold
    for sku in analysed_orders:

        period = 1
        order_receipt_index = {}
        final_stock = 0
        sim_window_collection = {}
        order_receipt_index = {}

        # create the sim_window for each sku, suing the random normal demand generated
        for i in range(0, period_length):

            po_qty_raised = 0
            #log.log(logging.INFO,'Current purchase order quantity {}'.format(po_qty_raised))

            # instantiate sim_window
            sim_window = MonteCarloWindow()

            # add sku_id
            sim_window.sku_id = sku.sku_id
            #log.log(logging.INFO,'Current SKU: {}'.format(sim_window.sku_id))
            # add closing stock
            previous_closing_stock = final_stock

            # mark sim_window.position or period in analysis
            sim_window.position = period

            #log.log(logging.INFO,('Current window position {}'.format(sim_window.position)))

            # add average orders to opening_stock if first period else add closing stock
            sim_window.opening_stock = opening_stock(sku.reorder_level, sku.safety_stock, sim_window.position,previous_closing_stock ) #calculated ltd until put into analyse orders

            # add random demand
            demand = random_normal_demand[0][sku.sku_id][i][0]
            sim_window.demand = demand
            #log.log(logging.INFO,('Current window demand {}'.format(sim_window.demand)))


            if sim_window.position in order_receipt_index.keys():
                sim_window.purchase_order_receipt_qty = order_receipt_index[sim_window.position]
                sim_window.po_number_received = 'PO {:.0f}{}'.format(sim_window.position, sim_window.index)
                del order_receipt_index[sim_window.position]
            else:
                sim_window.purchase_order_receipt_qty = 0
                sim_window.po_number_received = ''

            sim_window.index = index_item

            sim_window.backlog = backlog(opening_stock=sim_window.opening_stock,
                                         deliveries=sim_window.purchase_order_receipt_qty,
                                         demand=demand) + sim_window.previous_backlog

            sim_window.closing_stock = closing_stock(opening_stock=sim_window.opening_stock,
                                                     orders=demand,
                                                     deliveries=sim_window.purchase_order_receipt_qty,
                                                     backlog=sim_window.backlog)

            sim_window.holding_cost = holding_cost(sim_window.closing_stock, sku.unit_cost, holding_cost_percentage)

            sim_window.shortage_units = shortages(opening_stock=sim_window.opening_stock,
                                                  orders=demand,
                                                  deliveries=sim_window.purchase_order_receipt_qty)

            sim_window.shortage_cost = shortage_cost(closing_stock=(sim_window.backlog - sim_window.previous_backlog),
                                                     unit_cost=sku.unit_cost, shortage_cost_percentage=shortage_cost_percentage)

            sim_window.po_raised_flag = raise_po(reorder_level=sku.reorder_level, closing_stock=sim_window.closing_stock)

            po_receipt_period = period + sku.lead_time

            po_qty_raised = po_qty(eoq=sku.economic_order_qty,
                                   reorder_level=sku.reorder_level,
                                   backlog=sim_window.backlog,
                                   closing_stock=sim_window.closing_stock)

            if po_qty_raised > 0:
                order_receipt_index.update({po_receipt_period: po_qty_raised})
                sim_window.purchase_order_raised_qty = order_receipt_index.get(po_receipt_period)
            else:
                sim_window.purchase_order_raised_qty = 0

            sim_window.po_number_raised = ''

            if int(sim_window.purchase_order_raised_qty) > 0:
                sim_window.po_number_raised = 'PO {:.0f}{}'.format(po_receipt_period, sim_window.index)
                del po_receipt_period

            final_stock = sim_window.closing_stock

            sim_window.previous_backlog =  previous_backlog(sim_window.backlog,sim_window.closing_stock, sim_window.previous_backlog)

            sold = units_sold(sim_window.backlog, sim_window.opening_stock,
                                          sim_window.purchase_order_receipt_qty, sim_window.demand)
            sim_window.sold = sold
            sim_window.revenue = revenue(sku.unit_cost, sold)
            yield sim_window

            del sim_window

            period += 1
        index_item += 1



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging

from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import func
from sqlalchemy import join
from sqlalchemy import select
from supplychainpy._helpers._db_connection import engine
from supplychainpy._helpers._decorators import log_this

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

@log_this(logging.INFO, message='Retrieving Master SKU list')
def master_sku_list(uri: str) -> list:
    """ Connects to database to retrieve master sku list.

    Args:
        uri (str): Database connection string.

    Returns:
        list:   All unique SKU identification.

    """

    meta = MetaData()
    connection = engine(uri)

    try:
        log.log(logging.INFO,"Retrieving master sku list from database using uri connection string: \n {}.".format(uri))
        msk_table = Table('master_sku_list', meta, autoload=True, autoload_with=connection)
        skus = select([msk_table.columns.id, msk_table.columns.sku_id])
        rp = connection.execute(skus)
        result = []
        for i in rp:
            result.append((i['id'], i['sku_id']))
        log.log(logging.INFO, "Retrieval SUCCESSFUL. Number of records retrieved: {} .".format(len(result)))
    except OSError as e:
        print("Master Sku List retrieval failed. {}".format(e))
    else:
        rp.close()
        return result


def excess_controller(uri: str, direction: str = None, sku_id: str = None) -> tuple:
    """ Retrieves the excess SKUs in the inventory analysis table.

    Args:
        uri (str):          Database connection string.
        direction (str):    Indication of sort direction.
        sku_id (str):       SKU unique id.

    Returns:
        tuple:              Result.

    """
    meta = MetaData()
    connection = engine(uri)
    try:

        inventory_analysis = Table('inventory_analysis', meta, autoload=True, autoload_with=connection)
        if direction == 'biggest':
            skus = select([inventory_analysis.columns.sku_id, inventory_analysis.columns.excess_cost,
                           func.min(inventory_analysis.columns.excess_rank)])
        else:
            skus = select([inventory_analysis.columns.sku_id, inventory_analysis.columns.excess_cost,
                           func.max(inventory_analysis.columns.excess_rank)])

        # for i in inventory_analysis.columns:
        #    print(i)
        rp = connection.execute(skus)
        result = []
        sku_id = ""
        for i in rp:
            sku_id = str(i['sku_id'])
            result.append((i['sku_id'], i['excess_cost']))
        rp.close()
        # print(sku_id)
        msk_table = Table('master_sku_list', meta, autoload=True, autoload_with=connection)
        skus = select([msk_table.columns.id, msk_table.columns.sku_id]).where(msk_table.columns.id == sku_id)
        rp = connection.execute(skus
                                )

        for i in rp:
            result.append(i['sku_id'])
    except OSError as e:
        print("Retrieving {} excess failed. {}".format(direction, e))
    else:
        rp.close()
        return tuple(result)


def shortage_controller(uri: str, direction: str = None, sku_id: str = None) -> tuple:
    """ Retrieves SKUs with shortages.

    Args:
        uri (str):          Database connection string.
        direction (str):    Indication of sort direction.
        sku_id (str):       SKU unique id.

    Returns:
        tuple:              Result.


    """
    meta = MetaData()
    connection = engine(uri)
    inventory_analysis = Table('inventory_analysis', meta, autoload=True, autoload_with=connection)

    if direction == 'biggest':
        skus = select([inventory_analysis.columns.sku_id, inventory_analysis.columns.shortage_cost,
                       func.min(inventory_analysis.columns.shortage_rank)])
    else:
        skus = select([inventory_analysis.columns.sku_id, inventory_analysis.columns.shortage_cost,
                       func.max(inventory_analysis.columns.shortage_rank)])

    rp = connection.execute(skus)
    result = []
    sku_id = ""
    for i in rp:
        sku_id = str(i['sku_id'])
        result.append((i['sku_id'], i['shortage_cost']))
    rp.close()
    # print(sku_id)
    msk_table = Table('master_sku_list', meta, autoload=True, autoload_with=connection)
    skus = select([msk_table.columns.id, msk_table.columns.sku_id]).where(msk_table.columns.id == sku_id)
    rp = connection.execute(skus)
    for i in rp:
        result.append(i['sku_id'])
    rp.close()
    return tuple(result)


def revenue_controller(uri: str, direction: str = None, sku_id: str = None) -> tuple:
    """ Retrieves SKUs which generate revenue.

    Args:
        uri (str):          Database connection string.
        direction (str):    Indication of sort direction.
        sku_id (str):       SKU unique id.

    Returns:
        tuple:              Result.


    """
    meta = MetaData()
    connection = engine(uri)
    inventory_analysis = Table('inventory_analysis', meta, autoload=True, autoload_with=connection)

    if direction == 'smallest':
        skus = select([inventory_analysis.columns.sku_id, inventory_analysis.columns.revenue,
                       func.min(inventory_analysis.columns.revenue)])
    else:
        skus = select([inventory_analysis.columns.sku_id, inventory_analysis.columns.revenue,
                       func.max(inventory_analysis.columns.revenue)])

    rp = connection.execute(skus)
    result = []
    sku_id = ""
    for i in rp:
        sku_id = str(i['sku_id'])
        result.append((i['sku_id'], i['revenue']))
    rp.close()
    # print(sku_id)
    msk_table = Table('master_sku_list', meta, autoload=True, autoload_with=connection)
    skus = select([msk_table.columns.id, msk_table.columns.sku_id]).where(msk_table.columns.id == sku_id)
    rp = connection.execute(skus)
    for i in rp:
        result.append(i['sku_id'])
    rp.close()
    return tuple(result)


def inventory_turns_controller(uri: str, direction: str = None, sku_id: str = None) -> tuple:
    """ Retrieves SKU's Inventory Turns.

    Args:
        uri (str):          Database connection string.
        direction (str):    Indication of sort direction.
        sku_id (str):       SKU unique id.

    Returns:
        tuple:              Result.

    """
    meta = MetaData()
    connection = engine(uri)
    inventory_analysis = Table('inventory_analysis', meta, autoload=True, autoload_with=connection)
    if direction == 'smallest':
        skus = select([inventory_analysis.columns.sku_id, inventory_analysis.columns.inventory_turns,
                       func.min(inventory_analysis.columns.inventory_turns)])
    else:
        skus = select([inventory_analysis.columns.sku_id, inventory_analysis.columns.inventory_turns,
                       func.max(inventory_analysis.columns.inventory_turns)])

    rp = connection.execute(skus)
    result = []
    sku_id = ""
    for i in rp:
        sku_id = str(i['sku_id'])
        result.append((i['sku_id'], i['inventory_turns']))
    rp.close()
    # print(sku_id)
    msk_table = Table('master_sku_list', meta, autoload=True, autoload_with=connection)
    skus = select([msk_table.columns.id, msk_table.columns.sku_id]).where(msk_table.columns.id == sku_id)
    rp = connection.execute(skus)
    for i in rp:
        result.append(i['sku_id'])
    rp.close()
    return tuple(result)


def average_orders_controller(uri: str, direction: str = None, sku_id: str = None) -> tuple:
    """ Retrieves SKU's average orders.

    Args:
        uri (str):          Database connection string.
        direction (str):    Indication of sort direction.
        sku_id (str):       SKU unique id.

    Returns:
        tuple:              Result.

    """
    rp = None
    try:
        meta = MetaData()
        connection = engine(uri)
        inventory_analysis = Table('inventory_analysis', meta, autoload=True, autoload_with=connection)
        if direction == 'smallest':
            skus = select([inventory_analysis.columns.sku_id, inventory_analysis.columns.average_orders,
                           func.min(inventory_analysis.columns.average_orders)])
        else:
            skus = select([inventory_analysis.columns.sku_id, inventory_analysis.columns.average_orders,
                           func.max(inventory_analysis.columns.average_orders)])

        rp = connection.execute(skus)
        result = []
        sku_id = ""
        for i in rp:
            sku_id = str(i['sku_id'])
            result.append((i['sku_id'], i['average_orders']))
        rp.close()
        # print(sku_id)
        msk_table = Table('master_sku_list', meta, autoload=True, autoload_with=connection)
        skus = select([msk_table.columns.id, msk_table.columns.sku_id]).where(msk_table.columns.id == sku_id)
        rp = connection.execute(skus)
        for i in rp:
            result.append(i['sku_id'])

        return tuple(result)
    finally:
        rp.close()


def currency_symbol_controller(uri: str) -> str:
    """ Retrieves currency code from analysis database.

    Args:
        uri (str):          Database connection string.

    Returns:
        tuple:              Result.

    """
    meta = MetaData()
    connection = engine(uri)
    inventory_analysis = Table('inventory_analysis', meta, autoload=True, autoload_with=connection)
    transaction_log = Table('transaction_log', meta, autoload=True, autoload_with=connection)
    most_recent_transaction = select([transaction_log.columns.id, func.max(transaction_log.columns.date)]).limit(1)
    rp = connection.execute(most_recent_transaction)
    transaction_id = 0
    for i in rp:
        transaction_id = i['id']
    symbol_id = select([inventory_analysis.columns.currency_id]).where(
        inventory_analysis.columns.transaction_log_id == transaction_id).limit(1)
    rp.close()
    rp = connection.execute(symbol_id)
    currency_symbol_id = 0
    for i in rp:
        currency_symbol_id = i['currency_id']
    rp.close()
    currency = Table('currency', meta, autoload=True, autoload_with=connection)
    currency_symbol = select([currency.columns.currency_code]).where(currency.columns.id == currency_symbol_id)
    rp = connection.execute(currency_symbol)
    currency_code = 0
    for i in rp:
        currency_code = i['currency_code']
    rp.close()
    return currency_code


def classification_controller(uri: str, sku_id: str = None) -> tuple:
    """ Retrieves inventory classification for SKU.

    Args:
        uri (str):      Database connection string.
        sku_id (str):   SKU unique identification.

    Returns:
        tuple:  Result.

    """
    meta = MetaData()
    connection = engine(uri)
    inventory_analysis = Table('inventory_analysis', meta, autoload=True, autoload_with=connection)
    msk = Table('master_sku_list', meta, autoload=True, autoload_with=connection)
    j = join(inventory_analysis, msk, msk.columns.id == inventory_analysis.columns.sku_id)
    sku_classification = select([inventory_analysis.columns.abc_xyz_classification]).select_from(j).where(
        msk.columns.sku_id == sku_id)
    rp = connection.execute(sku_classification)
    classification = ''
    for i in rp:
        classification = i['abc_xyz_classification']

    return classification

    # if using html currency symbols think about how to deal with ascii when iteracting with dash on the command line etc


def safety_stock_controller(uri: str, sku_id: str = None, direction: str = None):
    """ Retrieves safety stock.

    Args:
        uri (str):          Database connection string.
        sku_id (str):       SKU unique identification.
        direction (str):    Indication of sort direction.

    Returns:

    """

    meta = MetaData()
    connection = engine(uri)
    inventory_analysis = Table('inventory_analysis', meta, autoload=True, autoload_with=connection)
    msk = Table('master_sku_list', meta, autoload=True, autoload_with=connection)
    j = join(inventory_analysis, msk, msk.columns.id == inventory_analysis.columns.sku_id)

    if direction == 'smallest':
        sku_classification = select([func.min(inventory_analysis.columns.safety_stock),
                                     msk.columns.sku_id]).select_from(j)
        rp = connection.execute(sku_classification)
    else:
        sku_classification = select([func.max(inventory_analysis.columns.safety_stock),
                                     msk.columns.sku_id]).select_from(j)

        rp = connection.execute(sku_classification)

    safety_stock = ''
    for i in rp:
        safety_stock = i[0]
        sku_identification = i[1]

    return safety_stock, sku_identification


def reorder_level_controller(uri: str, sku_id: str = None, direction: str = None):
    """ Retrieves reorder level.

    Args:
        uri:
        sku_id:
        direction:

    Returns:

    """
    meta = MetaData()
    connection = engine(uri)
    inventory_analysis = Table('inventory_analysis', meta, autoload=True, autoload_with=connection)
    msk = Table('master_sku_list', meta, autoload=True, autoload_with=connection)
    j = join(inventory_analysis, msk, msk.columns.id == inventory_analysis.columns.sku_id)

    if direction == 'smallest':
        sku_classification = select([func.min(inventory_analysis.columns.reorder_level),
                                     msk.columns.sku_id]).select_from(j)
        rp = connection.execute(sku_classification)
    else:
        sku_classification = select([func.max(inventory_analysis.columns.reorder_level),
                                     msk.columns.sku_id]).select_from(j)
        rp = connection.execute(sku_classification)

    reorder_level = ''
    sku_identification = ''
    for i in rp:
        reorder_level = i[0]
        sku_identification = i[1]

    return reorder_level, sku_identification



from decimal import Decimal


class Shortages:
    def __init__(self, unit_cost: Decimal, reorder_cost: Decimal, holding_cost: Decimal):
        self.__unit_cost = unit_cost
        self.__reorder_cost = reorder_cost
        self.__holding_cost = holding_cost

    @property
    def unit_cost(self) -> Decimal:
        return self.__unit_cost

    @unit_cost.setter
    def unit_cost(self, unit_cost:Decimal):
        self.__unit_cost = unit_cost

    @property
    def reorder_cost(self) -> Decimal:
        return self.__reorder_cost

    @reorder_cost.setter
    def reorder_cost(self, reorder_cost: Decimal):
        self.__reorder_cost = reorder_cost

    @property
    def holding_cost(self) ->Decimal:
        return self.__holding_cost

    @holding_cost.setter
    def holding_cost(self, holding_cost):
        self.__holding_cost = holding_cost

    def total_cost_per_cycle(self):
        pass

    def optimal_order_size(self):
        pass

    def optimal_back_order(self):
        pass



# Copyright (c) 2015-2016, The Authors and Contributors
# <see AUTHORS file>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
# following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

class ResponseBorg:
    """ Borg class making  class attributes global """
    _shared_response = {}  # global attribute dictionary remember to use single under score not double

    def __init__(self):
        self.__dict__ = self._shared_response


class ResponseSingleton(ResponseBorg):
    def __init__(self, **kwargs):
        # borg class instantiated at same time as the singleton updates the
        # shared_state dictionary by adding a new key-value pair
        ResponseBorg.__init__(self)
        self._shared_response.update(kwargs)

    def __str__(self):
        # returns the attribute for printing
        return str(self._shared_response)


class RecommendationStateMachine:
    def __init__(self):
        self.handlers = {}
        self.start_state = None
        self.end_states = []
        self.response = ResponseSingleton()

    def add_state(self, name, handler, end_state=0):
        name = name.upper()
        self.handlers[name] = handler
        if end_state:
            self.end_states.append(name)

    def set_start(self, name):
        self.start_state = name.upper()


class SkuMachine(RecommendationStateMachine):
    def __init__(self):
        super(SkuMachine, self).__init__()

    def run(self, sku_id):
        new_state = ''
        try:
            handler = self.handlers[self.start_state]
        except:
            raise OSError("call .set_start() first before .run()")
        if not self.end_states:
            raise OSError("must have at least one end_state")

        while True:
            (new_state, sku_id) = handler(sku_id)
            if new_state.upper() in self.end_states:
                break
            else:
                handler = self.handlers[new_state.upper()]
        return new_state, sku_id


class ProfileMachine(RecommendationStateMachine):
    def __init__(self):
        super(ProfileMachine, self).__init__()

    def run(self):
        new_state = ''
        try:
            handler = self.handlers[self.start_state]
        except:
            raise OSError("call .set_start() first before .run()")
        if not self.end_states:
            raise OSError("must have at least one end_state")

        while True:
            new_state = handler()
            if new_state.upper() in self.end_states:
                break
            else:
                handler = self.handlers[new_state.upper()]
        return new_state
